import 'dart:async';
import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'variables.dart';
import 'api.dart';
import 'dart:developer';
import 'function.dart';


class ChangeEvent {
  final EventType type;
  final Node? oldNode;
  final Node? newNode;

  ChangeEvent(this.type, this.oldNode, this.newNode);

  @override
  String toString() {
    return 'ChangeEvent{type: $type, oldNode: $oldNode, newNode: $newNode}';
  }
}


class _ConditionalCompleter {
  final bool Function(List<ChangeEvent>) canComplete;
  final List<ChangeEvent> Function(List<ChangeEvent>) createResult;

  final _completer = Completer<List<ChangeEvent>>();

  Future<List<ChangeEvent>> get future => _completer.future;

  _ConditionalCompleter(this.canComplete, this.createResult);

  void complete(List<ChangeEvent> events) {
    if (!canComplete(events)) return;
    _completer.complete(createResult(events));
  }
}


class SensorValueListener with StorageListener {
  List<ChangeEvent> events = <ChangeEvent>[];
  final Set<_ConditionalCompleter> completers = {};

  @override
  void gotStorageChange(EventType event, Node? oldNode, Node? newNode) async {
    //Event e = Event(event, oldNode, newNode);
    ChangeEvent e = ChangeEvent(event, oldNode, newNode);
    events.add(ChangeEvent(event, oldNode, newNode));
    for (final completer in completers) {
      completer.complete(events);
    }

    log('============================got event ${e.newNode!.path}');
    print('============================got event ${e.newNode!.path}');

    if(newNode!.path.contains(':data:metrics') && newNode.name != 'metrics'){

      events.add(e);
      Storage.getValuePath.add(newNode.parentPath);
      await updateData(events.iterator);
      events.remove(e);
    }else if(newNode.path.contains(':data:GeigerScoreUser')){

      events.add(e);
      var str = await newNode.getValue('implementedRecommendations');
      log('geiger-indicator - new IR: '+ str!.getValue('en')!);
      log('geiger-indicator - old IR: '+ CoreValues.uuidUserImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim());

      if(str.getValue('en') != CoreValues.uuidUserImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim()) {
        generateImplementedReco(str.getValue('en'),Types.users);
        await updateScoreNode(Types.users, Types.geigerScoreUser);
      }
      events.remove(e);

    }else if(newNode.path.contains(':data:GeigerScoreDevice') ){

      events.add(e);
      var str = await newNode.getValue('implementedRecommendations');
      log('geiger-indicator - new IR: '+ str!.getValue('en')!);
      log('geiger-indicator - old IR: '+ CoreValues.uuidDeviceImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim());

      if(str.getValue('en') != CoreValues.uuidDeviceImplementedRecommendations.toString().replaceAll('[','').replaceAll(']','').trim()){
        generateImplementedReco(str.getValue('en'),Types.devices);
        await updateScoreNode(Types.devices, Types.geigerScoreDevice);
      }
      events.remove(e);
    }

    //if(newNode?.name != 'metrics') await updateData(newNode!);



  }

//  Future<List<ChangeEvent>> getNumEvents(int num, [int timeout = 2000]) async {
//    int start = DateTime.now().millisecondsSinceEpoch;
//    List<ChangeEvent> ret = await Future.doWhile(() =>
//    events.length < num &&
//        start + 1000 * timeout > DateTime.now().millisecondsSinceEpoch)
//        .then((value) => events);
//
//    if (events.length < num) {
//      throw TimeoutException('Timeout reached while waiting for $num events');
//    }
//    return ret;
//  }

  Future<List<ChangeEvent>> awaitCount(int count,
      [Duration timeLimit = const Duration(seconds: 1)]) {
    canComplete(List events) => events.length >= count;
    createResult(List<ChangeEvent> events) =>
        List<ChangeEvent>.from(events.getRange(0, count));
    if (canComplete(events)) {
      return Future.value(createResult(events));
    }

    final completer = _ConditionalCompleter(canComplete, createResult);
    completers.add(completer);
    return completer.future.timeout(timeLimit, onTimeout: () {
      completers.remove(completer);
      throw TimeoutException(
          'Did not receive enough change events before timeout.', timeLimit);
    });
  }
}




Future updateData(Iterator<ChangeEvent> iterator) async{

  try {
    while (iterator.moveNext()) {
      var node = iterator.current.newNode;


      var temp = node
  !.path.toString().split(':');
  log('============>update');

  var type = '';
  temp[1] == Types.users ? type = Types.users : type = Types.devices;

  var type2 = '';
  temp[1] == Types.users ? type2 = Types.geigerScoreUser : type2 =
  Types.geigerScoreDevice;

  await updateMetrics(type, node);
  await updateScoreNode(type, type2);
  await updateRecommendationNode(type);

  await getDevices();
  await getEmployees();
  await crtAggregateScore();
  await updateScoreNode(Types.users, Types.
  geigerScoreAggregate
  );

}
}catch(e){
    log(e.toString());
    print(e);
  }

}


